/*
** 
**   IRC - Internet Relay Chat, commands/deny.c
**   Copyright (C) 2001 Adam Loghry
**
**   Portions Copyright (C) 1999 Chris Behrens (last 2 functions)
**   Portions Copyright (C) 2001 "srvx development team" (timechartype)
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**      
**   You should have received a copy of the GNU General Public License 
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/

#include "ircd_includes.h"
#include "commands.h"

RCSID("$Id: deny.c,v 1.1.1.1 2001/12/18 07:39:55 myr Exp $");

#define DEFAULT_DENY_REASON "No Reason"

static void add_deny(int enforce, char *oper, char *mask, time_t timeout, time_t when, int type, char *reason);
static void del_deny(IRCD_DENY *dptr);
static void _add_deny_to_list(IRCD_DENY **bucket, IRCD_DENY *dptr);
static void _del_deny_from_list(IRCD_DENY **bucket, IRCD_DENY *dptr);
static void deny_sync(IRCD_CLIENT *lptr, int type, char *args);
static IRCD_DENY *check_denied(char *mask, int type);
static void deny_enforce(IRCD_DENY *dptr);

int channel_isdenied(const char *channame)
{
	IRCD_DENY *dptr;

	for(dptr=Denies;dptr;dptr=dptr->next)
	{
		if(!(dptr->type & DENY_TYPE_CHANNEL) || !dptr->mask)
			continue;
		if(!strcasecmp(channame, dptr->mask))
			return 1; /* its denied */
	}

	return 0; /* its not */
}

int nick_isdenied(const char *nickname)
{
	IRCD_DENY *dptr;
	
	for(dptr=Denies;dptr;dptr=dptr->next)
	{
		if(!(dptr->type & DENY_TYPE_NICK) || !dptr->mask)
			continue;
		if(!match(dptr->mask, nickname))
			return 1;
	}

	return 0;
}

void check_deny_onuser(IRCD_CLIENT *cptr)
{
	IRCD_DENY 	*dptr;
	char		hostmask[USERLEN+HOSTLEN+1];
	int		islocal;
	int		isperm;

	sprintf(hostmask, "%s@%s", cptr->username, cptr->hostname);

	for(dptr=Denies;dptr;dptr=dptr->next)
	{
		if(!(dptr->type & DENY_TYPE_AKILL) || !dptr->mask)
			continue;

		islocal = (dptr->type & DENY_TYPE_LOCAL) ? 1 : 0;
		isperm  = (dptr->type & DENY_TYPE_PERM) ? 1 : 0;
		
		if(!match(dptr->mask, hostmask))
		{
			send_prefix_to_client(meptr, cptr, "NOTICE %s :*** You are not welcome on this %s; you have been autokilled.", 
								cptr->name, islocal ? "server" : "network");
			if(isperm)
				send_prefix_to_client(meptr, cptr, "NOTICE %s :*** Set on: %s by %s - does not expire (permanent) - Reason: %s",
						cptr->name, support_ctime(dptr->when), dptr->oper, dptr->reason);
			else
				send_prefix_to_client(meptr, cptr, "NOTICE %s :*** Set on: %s by %s - Expires on: %s - Reason: %s",
						cptr->name, support_ctime(dptr->when), dptr->oper, support_ctime(dptr->expire), dptr->reason);
			send_prefix_to_client(meptr, cptr, IRCD_REPLY(ERR_YOUREBANNEDCREEP, cptr->name), dptr->reason);

			return;
		}
	}
}

/* this function(and the code that calls it) ripped from srvx, http://srvx.net/ */
static long timechartype(char type)
{
    switch (type) {
                case 'y': return 365*24*60*60;
                case 'M': return 31*24*60*60;
                case 'w': return 7*24*60*60;
                case 'd': return 24*60*60;
                case 'h': return 60*60;
                case 'm': return 60;
                case 's': return 1;
                default: return 0;
        }
}

/* modified version of nickname_check() */
static int banned_nickname_check(char *nick)
{
        char    *ch;
	int 	isbad = 0;

        if ((*nick == '-') || isdigit(*nick))
                return 0;
        for (ch = nick; *ch && (ch - nick) < NICKLEN; ch++)
                if (!(((*ch) >= 'A' && (*ch) <= '~') || (*ch) == '*' || (*ch) == '?' || isdigit(*ch) || (*ch) == '-') || isspace(*ch))
                        isbad = 1;

        return (isbad) ? 0 : 1;
}


/* modified version of hostname_check() */
static int banned_hostname_check(char *hostname)
{
        char *tptr = hostname;
        char *dotptr = NULL;

        for(;*tptr;tptr++)
        {
		if ((*tptr == '*') || (*tptr == '?'))
			continue;
                if ((*tptr <= ' ') || (*tptr > 'z') || (*tptr == ':') || (*tptr == '%'))
                        break;
                if (*tptr == '.')
                        dotptr = tptr;
        }
        if (*tptr || !*hostname || !dotptr || (strlen(dotptr) > 4))
                return 0;

        return 1;
}

int m_gline(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

/* /deny [--(options)] +|-user@host|#chan|nick timeframe :reason
 *
 * channel bans are *NOT* recognized with wildcards(ie, you can deny #*lamers* and not be denying #asd123lamers-
 * nickname and akill denies are
 */

int m_deny(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	int 	add;
	int	type;
	char 	*mask;
	char	*expire;
	char	*reason;
	char	setter[NICKLEN*2]; /* safe length */

	add = type = 0;
	
        if(!IsReg(pptr) && !IsServer(lptr))
        {       
                 send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
                 return 3;
        }               
       
	if(MyClient(pptr))
	{
        	if(!IsAnOper(pptr))
        	{               
                	send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
                	return 3;
        	}                       
        	if(!(pptr->c_oline_flags & OLINE_FLAGS_CANDENY))
        	{       
        	        send_prefix_to_client(meptr, pptr, "NOTICE %s :You are not allowed to use DENY.", pptr->name);
        	        return 3;
        	}
	}

	mask = ircd_nextarg(&args, 0);

	if(mask)
	{
		/* a better solution here would be something like:
		 *
		 * while(!strncmp(mask, "--", 2))
		 * {
		 * 	....
		 * 	parse stuff
		 * 	....
		 * 	mask = ircd_nextarg(&args, 0);
		 * }
		 *
		 * right now, this is sufficient, since there can only be one flag per command. in the future, when --local is implemented, ill have to change that..
		 * 		--myriad */
		
		if(MyClient(pptr) && !strncmp(mask, "--", 2))
		{
			/* its /deny --something .. */

			if(!strcasecmp(mask, "--akill") || !strcasecmp(mask, "--gline") || !strcasecmp(mask, "--hostmask") ||
					!strcasecmp(mask, "--userhost"))
				type |= DENY_TYPE_AKILL;
			else if(!strcasecmp(mask, "--channel") || !strcasecmp(mask, "--channame") || !strcasecmp(mask, "--chanqline"))
				type |= DENY_TYPE_CHANNEL;
			else if(!strcasecmp(mask, "--nick") || !strcasecmp(mask, "--name") || !strcasecmp(mask, "--nickqline") || !strcasecmp(mask, "--nickname"))
				type |= DENY_TYPE_NICK;
			else if(!strcasecmp(mask, "--dump"))
			{
			        IRCD_DENY *dptr; 
                			for(dptr=Denies;(dptr && dptr->mask);dptr=dptr->next)
                        			send_prefix_to_client(meptr, pptr, "NOTICE %s :%s %d %d %c :(%s) %s",
                                                	pptr->name, dptr->mask, dptr->expire, dptr->when, 'X', dptr->oper, dptr->reason);
				return 3;
			}
			else
			{
				send_prefix_to_client(meptr, pptr, "NOTICE %s :Unknown DENY type", pptr->name);
				return 3;
			}

			mask = ircd_nextarg(&args, 0);
		}

		/* getting a deny burst from another server */
		if(IsServer(lptr) && !strncasecmp(mask, "!S", 2))
		{
			if(!strcasecmp(mask, "!SA"))
				type |= DENY_TYPE_AKILL;
			else if(!strcasecmp(mask, "!SC"))
				type |= DENY_TYPE_CHANNEL;
			else if(!strcasecmp(mask, "!SN"))
				type |= DENY_TYPE_NICK;

			if(type)
				deny_sync(lptr, type, args);

			return 3;
		}
		
		if(!MyClient(pptr))
		{
			if(!strcasecmp(mask, "!A"))
				type |= DENY_TYPE_AKILL;
			else if(!strcasecmp(mask, "!N"))
				type |= DENY_TYPE_NICK;
			else if(!strcasecmp(mask, "!C"))
				type |= DENY_TYPE_CHANNEL;
			else
				return 3; /* uhh.. */
			
			mask = ircd_nextarg(&args, 0);
		}

		/* okay, we have a mask without the "--crap" now 
		 * (or the !H/!N/!C stuff for clients on other servers */

		if(*mask == '+')
		{
			add = 1;
			mask++;
		}
		else if(*mask == '-')
		{
			add = 0;
			mask++;
		}
		else /* wont happen w/ remote clients */
			add = -1;

	}
	else
	{
		if(MyClient(pptr))
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "DENY");
		return 3;
	}

	expire = ircd_nextarg(&args, 0);
	reason = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);

	if(!mask || (add && (!expire || !reason)))
	{
		if(MyClient(pptr))
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "DENY");
		return 3;
	}

	if(!type) /* they didnt give us a type, so we'll try to guess one */
	{
		if(strchr(mask, '@') && strchr(mask, '.') && !(*mask == '#'))
		{
			if(!banned_hostname_check(mask))
			{
				if(MyClient(pptr))
					send_prefix_to_client(meptr, pptr, "NOTICE %s :Invalid hostmask to DENY.", pptr->name);
				return 3;
			}
			else
				type |= DENY_TYPE_AKILL;
		}
		else if(*mask == '#' || ((*mask == '&') && (type & DENY_TYPE_LOCAL)))
		{
			/* no idiot check performed here? */
			type |= DENY_TYPE_CHANNEL;
		}
		else if(banned_nickname_check(mask))
		{
			type |= DENY_TYPE_NICK;
		}
		else
		{
			if(MyClient(pptr))
				send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "DENY");
			return 3;
		}
	}
	else /* they -did- give us a type, so lets do idiot checks on it */
	{
		if(type & DENY_TYPE_NICK)
		{
			if(!banned_nickname_check(mask))
			{
				if(MyClient(pptr))
					send_prefix_to_client(meptr, pptr, "NOTICE %s :Bad nickname given for DENY", pptr->name);
				return 3;
			}
		}
		else if(type & DENY_TYPE_CHANNEL)
		{
			if(!(*mask == '#') && !((*mask == '&') && (type & DENY_TYPE_LOCAL)))
			{
				if(MyClient(pptr))
					send_prefix_to_client(meptr, pptr, "NOTICE %s :Bad channel name given for DENY", pptr->name);
				return 3;
			}
			if(vars_value_get_bynum(IRCD_VAR_WARN_CHANNEL_DENY) && (strchr(mask, '*') || strchr(mask, '?')))
				send_prefix_to_client(meptr, pptr, "NOTICE %s :Warning: channel names in DENY do not recognize wildcards(\"*\" or \"?\")", pptr->name);
		}
		else /* its a host. */
		{
			if(!banned_hostname_check(mask))
			{
				if(MyClient(pptr))
					send_prefix_to_client(meptr, pptr, "NOTICE %s :Bad hostname given for DENY", pptr->name);
				return 3;
			}
		}
	} /* done with forming the TYPE and its checks. */

	if(IsServer(pptr))
		strcpy(setter, pptr->name);
	else
		sprintf(setter, "%s[@%s]", pptr->name, pptr->servername);
	
	if(add)
	{
		unsigned long timeout = 0;
		/* parse expire */
		/* there's a disclaimer with this code; look at the top of the file. (in front of timechartype() -myriad */
		if(strchr(expire, 'y') || strchr(expire, 'M') || strchr(expire, 'w') ||
				strchr(expire, 'd') || strchr(expire, 'h') || strchr(expire, 'm') || strchr(expire, 's'))
		{
			char *interval = expire; /* we will want to display 'expire' to the opers later in our notice, so this is but a copy.. */
			unsigned long seconds = 0;
			int partial = 0;
			char c;
			while((c = *interval++))
			{
				if (isdigit((int)c)) 
				{
					partial = partial*10 + c - '0';
				}
				else
				{
					seconds += timechartype(c) * partial;
				}
			}

			timeout = seconds + partial + Now;
		}
		else if(!strcmp(expire, "P") || !strcmp(expire, "PERM") || !strcmp(expire, "PERMANENT"))
		{
			timeout = -1;
			type |= DENY_TYPE_PERM;
		}
		else
			timeout = atoi(expire);

		if(!(type & DENY_TYPE_PERM) && ((timeout <= 0) || (timeout <= Now)))
		{
			if(MyClient(pptr))
				send_prefix_to_client(meptr, pptr, "NOTICE %s :Error in expiration time! (%s(?))", pptr->name, support_ctime(timeout));
			return 3;
		}

		/* okay, we have a mask, we have a type, now.. */

		add_deny(1, pptr->name, mask, timeout, Now, type, (reason) ? reason : DEFAULT_DENY_REASON); 
	
		if(type & DENY_TYPE_NICK)
		{
			if(MyClient(pptr))
			{
				if(!(pptr->c_nflags & CLIENT_NFLAGS_DENIES))
					send_prefix_to_client(meptr, pptr, "WALLOPS :%s is adding a nickname DENY on: %s (%s)",
							setter, mask, expire);
			}
			send_wallops_to_flag(CLIENT_NFLAGS_DENIES, NULL, meptr, "%s is adding a nickname DENY on: %s (%s)",
					setter, mask, expire);
			send_to_servers(lptr, ":%s DENY !N +%s %s :%s",
								pptr->name, mask, expire, reason ? reason : DEFAULT_DENY_REASON);
		}
		else if(type & DENY_TYPE_CHANNEL)
		{
			if(MyClient(pptr))
			{
				if(!(pptr->c_nflags & CLIENT_NFLAGS_DENIES))
					send_prefix_to_client(meptr, pptr, "WALLOPS :%s is adding a channel DENY on: %s (%s)",
							setter, mask, expire);
			}
			send_wallops_to_flag(CLIENT_NFLAGS_DENIES, NULL, meptr, "%s is adding a channel DENY on: %s (%s)",
						setter, mask, expire);
			send_to_servers(lptr, ":%s DENY !C +%s %s :%s",
						pptr->name, mask, expire, reason ? reason : DEFAULT_DENY_REASON);
		}
		else if(type & DENY_TYPE_AKILL) /* u@h.. */
		{
			if(MyClient(pptr))
			{
				if(!(pptr->c_nflags & CLIENT_NFLAGS_DENIES))
					send_prefix_to_client(meptr, pptr, "WALLOPS :%s is adding an AKILL on: %s (%s)",
							 setter, mask, expire);
			}
			send_wallops_to_flag(CLIENT_NFLAGS_DENIES, NULL, meptr, "%s is adding an AKILL on: %s (%s)",
						setter, mask, expire);
			send_to_servers(lptr, ":%s DENY !A +%s %s :%s",
						pptr->name, mask, expire, reason ? reason : DEFAULT_DENY_REASON);
		}

		return 3;
	}
	else /* remove */
	{
		IRCD_DENY *dptr;
		int	isperm;
		int	islocal;

		isperm = islocal = 0;

		dptr = check_denied(mask, type);

		if(!dptr)
		{
			if(MyClient(pptr))
				send_prefix_to_client(meptr, pptr, "NOTICE %s :%s has no DENY", pptr->name, mask);
			return 3;
		}
		else
		{
#if 0
			if(dptr->mask)
				mask = dptr->mask;
#endif
			
			del_deny(dptr);

			if(type & DENY_TYPE_AKILL)
			{
				if(MyClient(pptr) && !(pptr->c_nflags & CLIENT_NFLAGS_DENIES))
					send_prefix_to_client(meptr, pptr, "WALLOPS :%s is removing AKILL on: %s", setter, mask);
				send_wallops_to_flag(CLIENT_NFLAGS_DENIES, NULL, meptr, "%s is removing AKILL on: %s", setter, mask);

				send_to_servers(lptr, ":%s DENY !A -%s", prefix, mask);
			}
			else if(type & DENY_TYPE_NICK)
			{
				if(MyClient(pptr) && !(pptr->c_nflags & CLIENT_NFLAGS_DENIES))
					send_prefix_to_client(meptr, pptr, "WALLOPS :%s is removing nickname DENY on: %s", setter, mask);
				send_wallops_to_flag(CLIENT_NFLAGS_DENIES, NULL, meptr, "%s is removing nickname DENY on: %s", setter, mask);

				send_to_servers(lptr, ":%s DENY !C -%s", prefix, mask);
			}
			else if(type & DENY_TYPE_CHANNEL)
			{
				if(MyClient(pptr) && !(pptr->c_nflags & CLIENT_NFLAGS_DENIES))
					send_prefix_to_client(meptr, pptr, "WALLOPS :%s is removing channel DENY on: %s", setter, mask);
				send_wallops_to_flag(CLIENT_NFLAGS_DENIES, NULL, meptr, "%s is removing channel DENY on: %s", setter, mask);
			}
			/* there is no else; we're done */

		}

		return 3;
	}

	return 3;

}

static void deny_enforce(IRCD_DENY *dptr)
{
	int islocal = (dptr->type & DENY_TYPE_LOCAL) ? 1 : 0;
	int isperm  = (dptr->type & DENY_TYPE_PERM) ? 1 : 0;

	if(dptr->type & DENY_TYPE_AKILL)
	{
		IRCD_CLIENT 	*cptr;
		char		userhost[USERLEN+HOSTLEN+1];	

		for(cptr=Locals;cptr;cptr=cptr->c_lo_next)
		{
			if(!IsReg(cptr)  || IsServer(cptr) || IsAnOper(cptr) || IsService(cptr) || (cptr->c_class->flags & CLASS_FLAGS_BYPASSDENY))
				continue;

			sprintf(userhost, "%s@%s", cptr->username, cptr->hostname);

			if(!match(dptr->mask, userhost))
			{
				char	exitbuf[KILLLEN];
				send_prefix_to_client(meptr, cptr, "NOTICE %s :*** You are no longer welcome on this %s; you have been autokilled.",
								cptr->name, islocal ? "server" : "network");
				if(isperm)
					send_prefix_to_client(meptr, cptr, "NOTICE %s :*** Set on: %s by %s - does not expire (permanent) - Reason: %s",
										cptr->name, support_ctime(dptr->when), dptr->oper, dptr->reason);
				else
					send_prefix_to_client(meptr, cptr, "NOTICE %s :*** Set on: %s by %s - Expires on: %s - Reason: %s",
									cptr->name, support_ctime(dptr->when), dptr->oper, support_ctime(dptr->expire), dptr->reason);
				send_prefix_to_client(meptr, cptr, IRCD_REPLY(ERR_YOUREBANNEDCREEP, cptr->name), dptr->reason);
				snprintf(exitbuf, KILLLEN, "Autokilled: %s", dptr->reason);
				cptr->c_flags |= CLIENT_FLAGS_KILLED;
				client_del(meptr, cptr, exitbuf);
			}
		}

	} 
	else if(dptr->type & DENY_TYPE_NICK)
	{
		IRCD_CLIENT *cptr;

		for(cptr=Locals;cptr;cptr=cptr->c_lo_next)
		{
			if(!IsReg(cptr) || IsServer(cptr) || IsAnOper(cptr) || IsService(cptr) || (cptr->c_class->flags & CLASS_FLAGS_BYPASSNICKNAMEDENY))
				continue;

			if(!match(dptr->mask, cptr->name))
			{
				char	exitbuf[KILLLEN];
				send_prefix_to_client(meptr, cptr, "NOTICE %s :*** Your current nickname is no longer valid on this %s; it has been denied.",
									cptr->name, islocal ? "server" : "network");
				if(isperm)
					send_prefix_to_client(meptr, cptr, "NOTICE %s :*** Set on: %s by %s - does not expire (permanent) - Reason: %s",
									cptr->name, support_ctime(dptr->when), dptr->oper, dptr->reason);
				else
					send_prefix_to_client(meptr, cptr, "NOTICE %s :*** Set on: %s by %s - Expires on: %s - Reason: %s",
									cptr->name, support_ctime(dptr->when), dptr->oper, support_ctime(dptr->expire), dptr->reason);
				snprintf(exitbuf, KILLLEN, "Invalid Nickname");
				cptr->c_nflags |= CLIENT_FLAGS_KILLED;
				client_del(meptr, cptr, exitbuf);
				/* .. when they reconnect, it will disallow them use of their nickname. */
			}
		}
	}
	else if(dptr->type & DENY_TYPE_CHANNEL) /* NOT RECOGNIZING WILDCARDS */
	{
		IRCD_CLIENT		*cptr;
		IRCD_CHANNEL 		*chptr = hash_find_channel(dptr->mask);
		IRCD_CLIENTLIST		*clptr;
		IRCD_CHANNELLIST	*chlptr;
		char			kickbuf[KILLLEN];

		if(!chptr)
			return; /* no channel! */
		
		snprintf(kickbuf, KILLLEN, "This channel has been denied (\"%s\")", dptr->reason);

		for(clptr=chptr->members;clptr;clptr=clptr->next)
		{
			if(clptr->cptr)
				cptr = clptr->cptr;
			else
				continue;

			if(MyClient(cptr))
			{
				if(IsAnOper(clptr->cptr) || (clptr->cptr->c_class->flags & CLASS_FLAGS_BYPASSCHANNELDENY) || IsService(clptr->cptr))
					continue; /* dont kick opers or ppl with 'bypasschanneldeny' in their class */
				/* these are sent as /NOTICE #chan :hi  -- they are only sent to local clients though. */
				send_prefix_to_client(meptr, clptr->cptr, "NOTICE %s :*** This channel is no longer valid on this %s; it has been denied.",
									chptr->name, islocal ? "server" : "network");
				if(isperm)
					send_prefix_to_client(meptr, clptr->cptr, "NOTICE %s :*** Set on: %s by %s - does not expire (permanent) - Reason: %s",
									chptr->name, support_ctime(dptr->when), dptr->oper, dptr->reason);
				else
					send_prefix_to_client(meptr, clptr->cptr, "NOTICE %s :*** Set on: %s by %s - Expires on: %s - Reason: %s",
									chptr->name, support_ctime(dptr->when), dptr->oper, support_ctime(dptr->expire), dptr->reason);
			
				send_to_channel_butone(NULL, CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY, chptr, meptr, "KICK %s %s :%s", 
									chptr->name, clptr->cptr->name, kickbuf);
				// send_prefix_to_client(meptr, cptr, ":%s KICK %s %s :%s", islocal ? meptr->name : NetHost, chptr->name, cptr->name, kickbuf);

				if(islocal) /* if its local, elegantly tell the other servers the guy parted */
					send_to_servers(NULL, ":%s PART %s", clptr->cptr->name, chptr->name);

				chlptr = IsMember(chptr, clptr->cptr); /* blech. */
				_channel_del_user(chptr, clptr->cptr, chlptr);
				del_channellist_from_list(&(clptr->cptr->c_channels), chlptr);
				MyFree(chlptr);
			}
		}
	}

	/* done */
}
		
static IRCD_DENY *check_denied(char *mask, int type)
{
	IRCD_DENY	*dptr;

	if(!type || !Denies)
		return NULL;

	for(dptr=Denies;dptr;dptr=dptr->next)
	{
		if(!dptr || !dptr->mask || !dptr->type)
			continue;
		if(!match(dptr->mask, mask) && (dptr->type & type))
			return dptr;
	}

	return NULL; /* didnt find a thing! */
}

/* used *only* in sync'ing a deny with the network during bursts. performs somewhat minimal sanity checks on bans to promote
 * a sync'd network. bans are being added(there wont be a sync to remove a deny) */
/* since we only get here if we are syncing bans, we dont potentially flood our opers with 100s of denies */
static void deny_sync(IRCD_CLIENT *lptr, int type, char *args)
{
	/* by now, we have in args: +(mask) (expire) (set-date) (setter) :(reason) */
	char	*mask 		= ircd_nextarg(&args, 0);
	char	*expire2 	= ircd_nextarg(&args, 0);
	char	*set2		= ircd_nextarg(&args, 0);
	char	*oper		= ircd_nextarg(&args, 0);
	char	*why		= ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	time_t  expire;
	time_t	set;

	if(expire2)
		expire = atoi(expire2);
	if(set2)
		set = atoi(set2);
	if(!mask || !set || !expire || !oper || !why)
		return; /* non-sane sync. */
	
	if(expire <= Now)
		return; /* the gline either already expired, or the sync is bad */

#if 0
	if(!nickname_check(oper))
		return;
#endif
	/* okay, everything appears to be in order */

	if(type & DENY_TYPE_AKILL)
	{
		if(banned_hostname_check(mask))
			add_deny(1, oper, mask, expire, set, type, why);
	}
	else if(type & DENY_TYPE_CHANNEL)
	{
		if(*mask == '#')
			add_deny(1, oper, mask, expire, set, type, why);
	}
	else if(type & DENY_TYPE_NICK)
	{
		if(banned_nickname_check(mask))
			add_deny(1, oper, mask, expire, set, type, why);
	}
}

static void _add_deny_to_list(IRCD_DENY **bucket, IRCD_DENY *dptr)
{
	dptr->last = NULL;
	if((dptr->next = *bucket) != NULL)
		dptr->next->last = dptr;
	*bucket = dptr;
}

static void _del_deny_from_list(IRCD_DENY **bucket, IRCD_DENY *dptr)
{
	if(dptr->last)
		dptr->last->next = dptr->next;
	else
		*bucket = dptr->next;
	if(dptr->next)
		dptr->next->last = dptr->last;
}
		
static void add_deny(int enforce, char *oper, char *mask, time_t timeout, time_t when, int type, char *reason)
{
	IRCD_DENY 		*dptr;

	dptr = (IRCD_DENY *)MyMalloc(sizeof(IRCD_DENY));

	dptr->mask		= MyStrdup(mask);		/* deny: mask */
	dptr->reason		= MyStrdup(reason);		/* deny: why */
	dptr->oper		= MyStrdup(oper);		/* deny: oper who did it */
	dptr->expire		= timeout;			/* deny: how long */
	dptr->type		= type;				/* deny: type */
	dptr->when		= when;

#if 0
	dptr->last 		= NULL;
	if ((dptr->next = *Denies) != NULL)
		dptr->next->last = dptr;
	*Denies 		= dptr;
#else
	_add_deny_to_list(&Denies, dptr);
#endif
	if(enforce)
		deny_enforce(dptr);

}

static void del_deny(IRCD_DENY *dptr)
{
#if 0
	if(dptr->last)
		dptr->last->next = dptr->next;
	else
		*Denies = dptr->next;
#else
	_del_deny_from_list(&Denies, dptr);
#endif
	
	MyFree(dptr->mask);
	MyFree(dptr->reason);
	MyFree(dptr->oper);
	dptr->expire 	= 0;	/* "delete"? */
	dptr->type 	= 0;
	dptr->when	= 0;

	MyFree(dptr);
}
