/*
**   IRC - Internet Relay Chat, source/commands.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   Portions Copyright (C) 2001 Adam Loghry
** 
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"
#include "cmds.h"
#include "../lib/confparse/configparse.h"

RCSID("$Id: commands.c,v 1.20 2001/10/02 01:40:35 cbehrens Exp $");


static void _send_var(IRCD_CLIENT *pptr, IRCD_VAR *vptr);
static void _send_var_to_opers(IRCD_CLIENT *pptr, IRCD_VAR *vptr);
static void _send_servers_recurs(IRCD_CLIENT *lptr, IRCD_CLIENT *cptr);
static void _count_servers_and_users(IRCD_CLIENT *acptr, int *numservers, int *numusers);
static int do_netset(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *var, char *value);

/* /SET (*)some_variable yes/no/on/off/1/100/blah blah blah */
/* * == forces the change to be sent to network. */
int m_set(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char		*var = ircd_nextarg(&args, 0);
	IRCD_VAR	*vptr;
	char		*value = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	int		ret;
	int		netset = 0;

	if(IsServer(lptr))
		return do_netset(lptr, pptr, prefix, var, value);
	
	if (!IsAnOper(lptr))
	{
		if (MyClient(pptr))
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}
	
	if (!var || !*var)
	{
		for(vptr=vars_value_getall();vptr;vptr=vars_value_getnext(vptr))
		{
			_send_var(pptr, vptr);
		}
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Your SET level is: %i", pptr->name, pptr->c_oline_setlevel);
		send_prefix_to_client(meptr, pptr, "NOTICE %s :*** End of vars",
				pptr->name);
		return 3;
	}

	if(*var == '*')
	{
		netset = 1;
		var++;
	}
	
#ifdef NO_HUB
	if(!strncasecmp(var, "HUB", 3))
	{
		send_prefix_to_client(meptr, pptr, "NOTICE %s :*** This server may not be a HUB. (NO_HUB)", pptr->name);
		return 3;
	}
#endif
	
	if (!value || !*value)
	{
		vptr = vars_find_byname(var);
		if (!vptr)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** Unknown var: %s", pptr->name, var);
			return 3;
		}
		_send_var(pptr, vptr);
		return 3;
	}

	if(!(pptr->c_oline_flags & OLINE_FLAGS_CANLOCALSET) && !(pptr->c_oline_flags & OLINE_FLAGS_CANGLOBALSET))
	{
		send_prefix_to_client(meptr, pptr, "NOTICE %s :You are not able to use SET.", pptr->name);
		return 3;
	}
	
	vptr = vars_find_byname(var);

	if (!vptr)
	{
		send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** Unknown var: %s", pptr->name, var);
		return 3;
	}

	if(vptr->v_net != 0 && vptr->v_net != 1)
		vptr->v_net = 0; /* misconfigured vars.h, this is an idiot check */
	
	if(!(pptr->c_oline_flags & OLINE_FLAGS_CANLOCALSET) && vptr->v_net == 0)
	{
		 send_prefix_to_client(meptr, pptr, "NOTICE %s :You are not allowed to locally set values! (%s is local)", pptr->name, vptr->v_name);
		 return 3;
	}

	if(!(pptr->c_oline_flags & OLINE_FLAGS_CANGLOBALSET) && vptr->v_net == 1)
	{
		send_prefix_to_client(meptr, pptr, "NOTICE %s :You are not allowed to globally set values! (%s is global)", pptr->name, vptr->v_name);
		return 3;
	}
	
	if(pptr->c_oline_setlevel < vptr->v_level)
	{
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Your SET level is not high enough to alter that value. (you: %i) (it: %i)",
						pptr->name, pptr->c_oline_setlevel, vptr->v_level);
		return 3;
	}

	if(vptr->v_net == 1)
		netset = 1;

	if (vptr->v_type == IRCD_VAR_TYPE_INT)
		value = (void *)atoi(value);
	else if (vptr->v_type == IRCD_VAR_TYPE_BOOL)
	{
		if (!value || !*value || !strncasecmp(value, "yes", 3) || !strncasecmp(value, "on", 2))
			value = (void *)1;
		else if (!strncasecmp(value, "no", 2) || !strncasecmp(value, "off", 3))
			value = (void *)0;
		else
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** SET failed: %s can only be set to YES/ON or NO/OFF",
					pptr->name, vptr->v_name);
			return 3;
		}
	}

	ret = vars_value_set_byname(pptr, var, value);

	if (ret)
	{
		send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** SET failed: %i (%s)", pptr->name, -ret,
				support_strerror(-ret));
		return 3;
	}

	_send_var(pptr, vptr);
	_send_var_to_opers(pptr, vptr);

	if(!ret && !strncasecmp(vptr->v_name, "HUB", 3))
		send_to_servers(NULL, ":%s WALLOPS :%s[@%s] has made %s a hub server!", meptr->name, pptr->name, meptr->name);
	
	if(netset && !ret && strncasecmp(vptr->v_name, "HUB", 3))
	{
		if(vptr->v_type == IRCD_VAR_TYPE_INT)
			send_to_servers(NULL, ":%s SET %s %i", pptr->name, vptr->v_name, vptr->v_value);
		else if(vptr->v_type == IRCD_VAR_TYPE_BOOL)
			send_to_servers(NULL, ":%s SET %s %i", pptr->name, vptr->v_name, vptr->v_value);
		else if(vptr->v_type == IRCD_VAR_TYPE_STRING)
			send_to_servers(NULL, ":%s SET %s %s", pptr->name, vptr->v_name, vptr->v_value);
	}
	
	return 3;
}

static void _send_var(IRCD_CLIENT *pptr, IRCD_VAR *vptr)
{
	switch(vptr->v_type)
	{
		case IRCD_VAR_TYPE_STRING:
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :[%i] %s%s set to %s",
				pptr->name, vptr->v_level, 
				(vptr->v_net) ? "*" : "", vptr->v_name,
				vptr->v_value ? vptr->v_value : "<null>");
			break;
		case IRCD_VAR_TYPE_INT:
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :[%i] %s%s set to %i", pptr->name,
				vptr->v_level, 
				(vptr->v_net) ? "*" : "", vptr->v_name, (int)vptr->v_value);
			break;
		case IRCD_VAR_TYPE_BOOL:
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :[%i] %s%s set to %s", pptr->name,
				vptr->v_level, 
				(vptr->v_net) ? "*" : "", vptr->v_name, vptr->v_value ? "YES" : "NO");
			break;
	}
}

static void _send_var_to_opers(IRCD_CLIENT *pptr, IRCD_VAR *vptr)
{
	switch(vptr->v_type)
	{
		case IRCD_VAR_TYPE_STRING:
			send_to_opers("%s SET %s%s(%i) to %s",
				pptr->name, (vptr->v_net) ? "*" : "", vptr->v_name,
				vptr->v_level, vptr->v_value ? vptr->v_value : "<null>");
			break;
		case IRCD_VAR_TYPE_INT:
			send_to_opers("%s SET %s%s(%i) to %i", pptr->name,
				(vptr->v_net) ? "*" : "", vptr->v_name, vptr->v_level, (int)vptr->v_value);
			break;
		case IRCD_VAR_TYPE_BOOL:
			send_to_opers("%s SET %s%s(%i) to %s", pptr->name,
				(vptr->v_net) ? "*" : "", vptr->v_name, vptr->v_level, vptr->v_value ? "YES" : "NO");
			break;
	}
}

static int do_netset(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *var, char *value)
{
	IRCD_VAR 	*vptr;
	int		ret;

	if(!var || !*var || !value || !*value)
		return 3;
	
	if(!IsAnOper(pptr))
		return 3;

	vptr = vars_find_byname(var);
	
	if(!vptr)
	{
		send_to_opers("Unable to perform network-based SET from %s[@%s]! (var \"%s\" does not exist!)",
					pptr->name, pptr->servername, var);
		return 3;
	}

	if (vptr->v_type == IRCD_VAR_TYPE_INT)
		value = (void *)atoi(value);
	else if (vptr->v_type == IRCD_VAR_TYPE_BOOL)
	{
		if ((int)atoi(value) == 1)
			value = (void *)1;
		else if ((int)atoi(value) == 0)
			value = (void *)0;
		else
		{
			send_to_opers("Unable to perform network-based SET from %s[@%s]! (%s cannot be \"%s\")",
						pptr->name, pptr->servername, vptr->v_name, value);
			return 3;
		}
	}

	ret = vars_value_set_byname(pptr, var, value);

	if (ret)
	{
		send_to_opers("Unable to perform network-based SET from %s[@%s]! (%i:%s)", pptr->name, pptr->servername, -ret, support_strerror(-ret));
		return 3;
	}

#if 0
	_send_var_to_opers(pptr, vptr);
#endif

	if(vptr->v_type == IRCD_VAR_TYPE_INT)
	{
		send_to_opers("%s[@%s] has set network-wide variable %s to: %i",
					pptr->name, pptr->servername, vptr->v_name, vptr->v_value);
		send_to_servers(lptr, ":%s SET %s %i", 
					prefix, vptr->v_name, vptr->v_value);
	} else if (vptr->v_type == IRCD_VAR_TYPE_BOOL) {
		send_to_opers("%s[@%s] has set network-wide variable %s to: %s", 
					pptr->name, pptr->servername, vptr->v_name,
					((int)vptr->v_value == 1) ? "YES" : "NO");
		send_to_servers(lptr, ":%s SET %s %i",
					prefix, vptr->v_name, vptr->v_value);
	} else if (vptr->v_type == IRCD_VAR_TYPE_STRING) {
		send_to_opers("%s[@%s] has set network-wide variable %s to: %s",
					pptr->name, pptr->servername, vptr->v_name, vptr->v_value);
		send_to_servers(lptr, ":%s SET %s %s", prefix, vptr->v_name, vptr->v_value);
	}

	return 3;
}
